# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gzip
import numpy as np
import signal
import os
import warnings
import datetime
import h5py
from multiprocessing import Pool
from functools import partial

# globals
__SPECTROGRAPH_EXPECTED_HEIGHT = 1024
__SPECTROGRAPH_EXPECTED_WIDTH = 256
__SPECTROGRAPH_RAW_DT = np.dtype("uint16")
__SPECTROGRAPH_RAW_DT = __SPECTROGRAPH_RAW_DT.newbyteorder('>')  # force big endian byte ordering
__SPECTROGRAPH_PROCESSED_DT = np.dtype("float32")


def read_raw(file_list, n_parallel=1, first_record=False, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # throw up warning if the no_metadata flag is true and start/end times supplied
    #
    # NOTE: since the timestamp info is in the metadata, we can't filter based on
    # start or end time if we were told to not read the metadata, so we throw up
    # a warning and read all the data, and don't read the metadata (no_metadata being
    # true takes priority).
    if (no_metadata is True and (start_time is not None or end_time is not None)):
        warnings.warn("Cannot filter on start or end time if the no_metadata parameter is set to True. Set no_metadata to False " +
                      "to allow filtering on times. Will proceed by skipping filtering on times and returning no metadata.",
                      UserWarning,
                      stacklevel=1)
        start_time = None
        end_time = None

    # check n_parallel
    if (n_parallel > 1):
        try:
            # set up process pool (ignore SIGINT before spawning pool so child processes inherit SIGINT handler)
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool = Pool(processes=n_parallel)
            signal.signal(signal.SIGINT, original_sigint_handler)  # restore SIGINT handler
        except ValueError:  # pragma: nocover
            # likely the read call is being used within a context that doesn't support the usage
            # of signals in this way, proceed without it
            pool = Pool(processes=n_parallel)

        # call readfile function, run each iteration with a single input file from file_list
        # NOTE: structure of data - data[file][metadata dictionary lists = 1, images = 0][frame]
        data = []
        try:
            data = pool.map(
                partial(
                    __spectrograph_raw_readfile_worker,
                    first_record=first_record,
                    no_metadata=no_metadata,
                    start_time=start_time,
                    end_time=end_time,
                    quiet=quiet,
                ), file_list)
        except KeyboardInterrupt:  # pragma: nocover
            pool.terminate()  # gracefully kill children
            return np.empty((0, 0, 0), dtype=__SPECTROGRAPH_RAW_DT), [], []
        else:
            pool.close()
            pool.join()
    else:
        # don't bother using multiprocessing with one worker, just call the worker function directly
        data = []
        for f in file_list:
            data.append(
                __spectrograph_raw_readfile_worker(
                    f,
                    first_record=first_record,
                    no_metadata=no_metadata,
                    start_time=start_time,
                    end_time=end_time,
                    quiet=quiet,
                ))

    # derive number of frames to prepare for
    total_num_frames = 0
    image_width = __SPECTROGRAPH_EXPECTED_WIDTH
    image_height = __SPECTROGRAPH_EXPECTED_HEIGHT
    for i in range(0, len(data)):
        if (data[i][2] is True):
            continue
        if (len(data[i][0]) != 0):
            total_num_frames += data[i][0].shape[2]
            image_height = data[i][0].shape[0]
            image_width = data[i][0].shape[1]

    # pre-allocate array sizes
    images = np.empty([image_height, image_width, total_num_frames], dtype=__SPECTROGRAPH_RAW_DT)
    if (no_metadata is False):
        metadata_dict_list = [{}] * total_num_frames
    else:
        metadata_dict_list = []
    problematic_file_list = []

    # populate data
    list_position = 0
    for i in range(0, len(data)):
        # check if file was problematic
        if (data[i][2] is True):
            problematic_file_list.append({
                "filename": data[i][3],
                "error_message": data[i][4],
            })
            continue

        # check if any data was read in
        if (len(data[i][0]) == 0):
            continue

        # find actual number of frames, this may differ from predicted due to dropped frames, end
        # or start of imaging
        real_num_frames = data[i][0].shape[2]

        # metadata dictionary list at data[][1]
        if (no_metadata is False):
            metadata_dict_list[list_position:list_position + real_num_frames] = data[i][1]
        images[:, :, list_position:list_position + real_num_frames] = data[i][0]  # image arrays at data[][0]
        list_position = list_position + real_num_frames  # advance list position

    # trim unused elements from predicted array sizes
    if (no_metadata is False):
        metadata_dict_list = metadata_dict_list[0:list_position]
    images = np.delete(images, range(list_position, total_num_frames), axis=2)

    # ensure entire array views as uint16
    images = images.astype(np.uint16)

    # return
    data = None
    return images, metadata_dict_list, problematic_file_list


def __spectrograph_raw_readfile_worker(file, first_record=False, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # init
    images = np.array([])
    metadata_dict_list = []
    is_first = True
    metadata_dict = {}
    site_uid = ""
    device_uid = ""
    problematic = False
    error_message = ""

    # convert to str to handle path type
    file = str(file)

    # extract start and end times of the filename
    #
    # NOTE: we use this to better inform us about zero-filesize files
    try:
        file_dt = datetime.datetime.strptime(os.path.basename(file)[0:13], "%Y%m%d_%H%M")
    except Exception:
        if (quiet is False):
            print("Failed to extract timestamp from filename")
        problematic = True
        error_message = "failed to extract timestamp from filename"
        return images, metadata_dict_list, problematic, file, error_message

    # set site UID and device UID in case we need it (ie. dark frames, or unstacked files)
    file_split = os.path.basename(file).split('_')
    if (len(file_split) == 5):
        # is a regular file
        site_uid = file_split[2]
        device_uid = file_split[3]
    elif (len(file_split) > 5):
        # is likely a dark frame or a unstacked frame
        site_uid = file_split[3]
        device_uid = file_split[4]

    # check file extension to see if it's gzipped or not
    unzipped = None
    try:
        if file.endswith("pgm.gz"):
            unzipped = gzip.open(file, mode='rb')
        elif file.endswith("pgm"):
            unzipped = open(file, mode='rb')
        else:
            if (quiet is False):
                print("Unrecognized file type: %s" % (file))
            problematic = True
            error_message = "Unrecognized file type"
            try:
                if (unzipped is not None):
                    unzipped.close()
            except Exception:  # pragma: nocover
                pass
            return images, metadata_dict_list, problematic, file, error_message
    except Exception as e:
        if (quiet is False):
            print("Failed to open file '%s' " % (file))
        problematic = True
        error_message = "failed to open file: %s" % (str(e))
        try:
            if (unzipped is not None):
                unzipped.close()
        except Exception:  # pragma: nocover
            pass
        return images, metadata_dict_list, problematic, file, error_message

    # read the file
    prev_line = None
    line = None
    skip_this_record = False
    while True:
        # break out depending on first_record param
        if (first_record is True and is_first is False):
            break

        # read a line
        try:
            prev_line = line
            line = unzipped.readline()
        except Exception as e:
            if (quiet is False):
                print("Error reading before image data in file '%s'" % (file))
            problematic = True
            metadata_dict_list = []
            images = np.array([])
            error_message = "error reading before image data: %s" % (str(e))
            try:
                unzipped.close()
            except Exception:  # pragma: nocover
                pass
            return images, metadata_dict_list, problematic, file, error_message

        # break loop at end of file
        if (line == b''):
            break

        # magic number; this is not a metadata or image line, exclude
        if (line.startswith(b'P5\n')):  # type: ignore
            continue

        # process line
        if (line.startswith(b'#"')):  # type: ignore
            if (no_metadata is False):
                # metadata lines start with #"<key>"
                try:
                    line_decoded = line.decode("ascii")  # type: ignore
                except Exception as e:
                    # skip metadata line if it can't be decoded, likely corrupt file
                    if (quiet is False):
                        print("Error decoding metadata line: %s (line='%s', file='%s')" % (str(e), line, file))
                    problematic = True
                    error_message = "error decoding metadata line: %s" % (str(e))
                    continue

                # split the key and value out of the metadata line
                line_decoded_split = line_decoded.split('"')
                key = line_decoded_split[1]
                value = line_decoded_split[2].strip()

                # add entry to dictionary
                metadata_dict[key] = value

                # set the site/device uids, or inject the site and device UIDs if they are missing
                if ("Site unique ID" not in metadata_dict):
                    metadata_dict["Site unique ID"] = site_uid
                else:
                    site_uid = metadata_dict["Site unique ID"]
                if ("Imager unique ID" not in metadata_dict):
                    metadata_dict["Imager unique ID"] = device_uid
                else:
                    device_uid = metadata_dict["Imager unique ID"]

                # split dictionaries up per frame, exposure plus initial readout is
                # always the end of metadata for frame
                if (key.startswith("Exposure plus readout")):
                    # check if we want to skip this frame based on the start and end times
                    this_timestamp = datetime.datetime.strptime(metadata_dict["Image request start"],
                                                                "%Y-%m-%d %H:%M:%S.%f UTC").replace(microsecond=0)
                    if ((start_time is None or this_timestamp >= start_time) and (end_time is None or this_timestamp <= end_time)):
                        metadata_dict_list.append(metadata_dict)
                    else:
                        skip_this_record = True
                    metadata_dict = {}
        elif line == b'65535\n':
            # there are 2 lines between "exposure plus read out" and the image
            # data, the first is the image dimensions and the second is the max
            # value
            #
            # check the previous line to get the dimensions of the image
            prev_line_split = prev_line.decode("ascii").strip().split()  # type: ignore
            image_width = int(prev_line_split[0])
            image_height = int(prev_line_split[1])
            bytes_to_read = image_width * image_height * 2  # 16-bit image depth

            # read image
            try:
                # skip the image data if we wanted to
                if (skip_this_record is True):
                    skip_this_record = False
                    unzipped.seek(bytes_to_read, 1)
                    continue

                # read the image size in bytes from the file
                image_bytes = unzipped.read(bytes_to_read)

                # format bytes into numpy array of unsigned shorts (2byte numbers, 0-65536),
                # effectively an array of pixel values
                image_np = np.frombuffer(image_bytes, dtype=__SPECTROGRAPH_RAW_DT)  # type: ignore

                # change 1d numpy array into matrix with correctly located pixels
                image_matrix = np.reshape(image_np, (image_height, image_width, 1))
            except Exception as e:
                if (quiet is False):
                    print("Failed reading image data frame: %s" % (str(e)))
                metadata_dict_list.pop()  # remove corresponding metadata entry
                problematic = True
                error_message = "image data read failure: %s" % (str(e))
                continue  # skip to next frame

            # initialize image stack
            if (is_first is True):
                images = image_matrix
                is_first = False
            else:
                images = np.dstack([images, image_matrix])  # depth stack images (on 3rd axis)

    # close gzip file
    unzipped.close()

    # check to see if the image is empty
    image_size_is_zero = False
    if (start_time is None and end_time is None):
        if (images.size == 0):
            image_size_is_zero = True
    elif (start_time is not None and end_time is not None):
        if (file_dt >= start_time and file_dt <= end_time):
            if (images.size == 0):
                image_size_is_zero = True
    elif (start_time is not None and file_dt >= start_time):
        if (images.size == 0):
            image_size_is_zero = True
    elif (end_time is not None and file_dt <= end_time):
        if (images.size == 0):
            image_size_is_zero = True
    if (image_size_is_zero is True):
        if (quiet is False):
            print("Error reading image file: found no image data")
        problematic = True
        error_message = "no image data"

    # return
    return images, metadata_dict_list, problematic, file, error_message


def __str_to_datetime_formatter(timestamp_str):
    return datetime.datetime.strptime(timestamp_str.decode(), "%Y-%m-%d %H:%M:%S UTC")


def read_processed(file_list, n_parallel=1, first_record=False, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # throw up warning if the no_metadata flag is true and start/end times supplied
    #
    # NOTE: since the timestamp info is in the metadata, we can't filter based on
    # start or end time if we were told to not read the metadata, so we throw up
    # a warning and read all the data, and don't read the metadata (no_metadata being
    # true takes priority).
    if (no_metadata is True and (start_time is not None or end_time is not None)):
        warnings.warn("Cannot filter on start or end time if the no_metadata parameter is set to True. Set no_metadata to False " +
                      "to allow filtering on times. Will proceed by skipping filtering on times and returning no metadata.",
                      UserWarning,
                      stacklevel=1)
        start_time = None
        end_time = None

    # check n_parallel
    if (n_parallel > 1):
        try:
            # set up process pool (ignore SIGINT before spawning pool so child processes inherit SIGINT handler)
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool = Pool(processes=n_parallel)
            signal.signal(signal.SIGINT, original_sigint_handler)  # restore SIGINT handler
        except ValueError:  # pragma: nocover
            # likely the read call is being used within a context that doesn't support the usage
            # of signals in this way, proceed without it
            pool = Pool(processes=n_parallel)

        # call readfile function, run each iteration with a single input file from file_list
        # NOTE: structure of data - data[file][metadata dictionary lists = 1, images = 0][frame]
        data = []
        try:
            data = pool.map(
                partial(
                    __spectrograph_processed_readfile_worker,
                    first_record=first_record,
                    no_metadata=no_metadata,
                    start_time=start_time,
                    end_time=end_time,
                    quiet=quiet,
                ), file_list)
        except KeyboardInterrupt:  # pragma: nocover
            pool.terminate()  # gracefully kill children
            return np.empty((0, 0, 0), dtype=__SPECTROGRAPH_PROCESSED_DT), np.empty((0), dtype=datetime.datetime), [], []
        else:
            pool.close()
            pool.join()
    else:
        # don't bother using multiprocessing with one worker, just call the worker function directly
        data = []
        for f in file_list:
            data.append(
                __spectrograph_processed_readfile_worker(
                    f,
                    first_record=first_record,
                    no_metadata=no_metadata,
                    start_time=start_time,
                    end_time=end_time,
                    quiet=quiet,
                ))

    # derive number of frames to prepare for
    total_num_frames = 0
    image_width = __SPECTROGRAPH_EXPECTED_WIDTH
    image_height = __SPECTROGRAPH_EXPECTED_HEIGHT
    for i in range(0, len(data)):
        if (data[i][3] is True):
            continue
        if (len(data[i][0]) != 0):  # type: ignore
            total_num_frames += data[i][0].shape[2]  # type: ignore
            image_height = data[i][0].shape[0]  # type: ignore
            image_width = data[i][0].shape[1]  # type: ignore

    # pre-allocate array sizes
    spectra = np.empty((image_height, image_width, total_num_frames), dtype=__SPECTROGRAPH_PROCESSED_DT)
    timestamp = np.empty((total_num_frames), dtype=object)
    if (no_metadata is False):
        metadata_dict_list = [{}] * total_num_frames
    else:
        metadata_dict_list = []
    problematic_file_list = []

    # populate data
    list_position = 0
    for i in range(0, len(data)):
        # check if file was problematic
        if (data[i][3] is True):
            problematic_file_list.append({
                "filename": data[i][4],
                "error_message": data[i][5],
            })
            continue

        # check if any data was read in
        if (len(data[i][0]) == 0):  # type: ignore
            continue

        # find actual number of frames, this may differ from predicted due to
        # dropped frames, end or start of imaging
        real_num_frames = data[i][0].shape[2]  # type: ignore

        # populate image data
        spectra[:, :, list_position:list_position + real_num_frames] = data[i][0]

        # populate timestamps
        timestamp[list_position:list_position + real_num_frames] = data[i][1]

        # populate metadata objects
        if (no_metadata is False):
            metadata_dict_list[list_position:list_position + real_num_frames] = data[i][2]

        # advance list position
        list_position = list_position + real_num_frames

    # trim unused elements from predicted array sizes
    if (no_metadata is False):
        metadata_dict_list = metadata_dict_list[0:list_position]
    spectra = np.delete(spectra, range(list_position, total_num_frames), axis=2)

    # ensure entire array views as float32
    spectra = spectra.astype(__SPECTROGRAPH_PROCESSED_DT)

    # convert timestamps to datetime objections
    timestamp = np.vectorize(__str_to_datetime_formatter)(timestamp)  # type: ignore
    timestamp = timestamp.astype(datetime.datetime)

    # return
    data = None
    return spectra, timestamp, metadata_dict_list, problematic_file_list


def __spectrograph_processed_readfile_worker(file, first_record=False, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # init
    spectra = np.array([])
    timestamps = np.array([])
    metadata_dict_list = []
    problematic = False
    error_message = ""

    # process file
    try:
        # open H5 file
        f = h5py.File(file, 'r')

        # get timestamps
        if (first_record is True):
            original_timestamps = [f["data"]["timestamp"][0]]  # type: ignore
        else:
            original_timestamps = f["data"]["timestamp"][:]  # type: ignore

        # search through timestamps for indexes to read data for, based
        # on the start and end times
        idxs = []
        timestamps = []
        if (start_time is None and end_time is None):
            # no filtering, read all indexes
            idxs = np.arange(0, len(original_timestamps))  # type: ignore
            timestamps = original_timestamps
        else:
            for i, val in enumerate(original_timestamps):  # type: ignore
                val_dt = datetime.datetime.strptime(val.decode(), "%Y-%m-%d %H:%M:%S UTC")
                if ((start_time is None or val_dt >= start_time) and (end_time is None or val_dt <= end_time)):
                    # matches start and end time range, we want to read this index
                    idxs.append(i)
                    timestamps.append(original_timestamps[i])  # type: ignore

        # bail out if we don't want to read any frames
        if (len(idxs) == 0):
            timestamps = np.array([])
            return spectra, timestamps, metadata_dict_list, problematic, file, error_message

        # get spectra
        spectra = f["data"]["spectra"][:, :, idxs]  # type: ignore

        # read metadata
        file_metadata = {}
        if (no_metadata is False):
            # get file metadata
            for key, value in f["metadata"]["file"].attrs.items():  # type: ignore
                file_metadata[key] = value

            # read wavelength
            file_metadata["wavelength"] = f["data"]["wavelength"][:]  # type: ignore
            metadata_dict_list = [file_metadata] * len(idxs)  # type: ignore

        # close H5 file
        f.close()

        # set image vars and reshape if multiple images
        image_height = spectra.shape[0]  # type: ignore
        image_width = spectra.shape[1]  # type: ignore
        if (len(spectra.shape) == 2):  # type: ignore
            # force reshape to 3 dimensions
            spectra = spectra.reshape((image_height, image_width, 1))  # type: ignore
    except Exception as e:
        if (quiet is False):
            print("Error reading image file: %s" % (str(e)))
        problematic = True
        error_message = "error reading image file: %s" % (str(e))

    # return
    return spectra, timestamps, metadata_dict_list, problematic, file, error_message



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import signal
import numpy as np
from pathlib import Path
from multiprocessing import Pool
from functools import partial
from ..classes import RiometerData

# globals
RIOMETER_DT = np.dtype("float32")
NORSTAR_RIOMETER_3_LETTER_SITE_CODES = {
    "chur": ["chu"],
    "cont": ["con"],
    "daws": ["daw"],
    "arvi": ["esk"],
    "fsim": ["sim"],
    "fsmi": ["smi"],
    "gill": ["gil"],
    "isll": ["isl"],
    "mcmu": ["mcm"],
    "pina": ["pin"],
    "rabb": ["rab"],
    "rank": ["ran"],
    "talo": ["tal"],
}  # NOTE: pulled from UCalgary SRS API code; if you update this, we should update the API too.


def read(file_list, n_parallel=1, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # if input is just a single file name in a string, convert to a list to be fed to the workers
    if isinstance(file_list, str) or isinstance(file_list, Path):
        file_list = [file_list]

    # check n_parallel
    if (n_parallel > 1):
        try:
            # set up process pool (ignore SIGINT before spawning pool so child processes inherit SIGINT handler)
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool = Pool(processes=n_parallel)
            signal.signal(signal.SIGINT, original_sigint_handler)  # restore SIGINT handler
        except ValueError:  # pragma: nocover
            # likely the read call is being used within a context that doesn't support the usage
            # of signals in this way, proceed without it
            pool = Pool(processes=n_parallel)

        # call readfile function, run each iteration with a single input file from file_list
        # NOTE: structure of data - data[file][metadata dictionary lists = 1, images = 0][frame]
        data = []
        try:
            data = pool.map(partial(
                __riometer_readfile_worker,
                no_metadata=no_metadata,
                start_time=start_time,
                end_time=end_time,
                quiet=quiet,
            ), file_list)
        except KeyboardInterrupt:  # pragma: nocover
            pool.terminate()  # gracefully kill children
            return [], [], [], []
        else:
            pool.close()
            pool.join()
    else:
        # don't bother using multiprocessing with one worker, just call the worker function directly
        data = []
        for f in file_list:
            data.append(__riometer_readfile_worker(
                f,
                no_metadata=no_metadata,
                start_time=start_time,
                end_time=end_time,
                quiet=quiet,
            ))

    # compile results
    rio_data = []
    top_level_timestamps = []
    metadata = []
    problematic_file_list = []
    for result in data:
        # problematic
        if (result["problematic"] is True):
            problematic_file_list.append({
                "filename": result["file"],
                "error_message": result["error_message"],
            })
            continue

        # riometer data
        rio_data.append(RiometerData(timestamp=result["np_timestamp"], raw_signal=result["np_raw_signal"], absorption=result["np_absorption"]))

        # timestamp
        if (len(result["np_timestamp"]) > 0):
            top_level_timestamps.append(result["np_timestamp"][0])

        # metadata
        if (no_metadata is False):
            metadata.append(result["metadata"])

    # return
    return rio_data, top_level_timestamps, metadata, problematic_file_list


def __riometer_readfile_worker(file, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # init
    metadata_dict = {}
    np_timestamp = np.array([], dtype=datetime.datetime)
    np_raw_signal = np.array([], dtype=RIOMETER_DT)
    np_absorption = np.array([], dtype=RIOMETER_DT)
    problematic = False
    error_message = ""

    # determine number of expected columns
    #
    # NOTE: we use the filename here to tell us if it's a k0 or k2 file, since
    # we cannot assume the dataset name will be supplied to the parent functions.
    file_type = None
    if ("_k0_" in file.name or "v0.txt" in file.name):
        file_type = "k0"
        np_absorption = None
    elif ("_k2_" in file.name or "v1a.txt" in file.name):
        file_type = "k2"

    # check file type
    if (file_type is None):
        if (quiet is False):
            print("Error reading file, unknown file type for '%s'" % (file))
        problematic = True
        error_message = "error reading file, unknown file type"
        return {
            "np_timestamp": np_timestamp,
            "np_raw_signal": np_raw_signal,
            "np_absorption": np_absorption,
            "metadata": metadata_dict,
            "problematic": problematic,
            "file": file,
            "error_message": error_message,
        }

    # read the file using numpy
    if (file_type == "k0"):
        # k0 data, 3 columns
        np_date, np_time, np_raw_signal = np.genfromtxt(file, comments='#', dtype="S8,S8,f", unpack=True)
    else:
        # k2 data, 4 columns
        np_date, np_time, np_raw_signal, np_absorption = np.genfromtxt(file, comments='#', dtype="S8,S8,f,f", unpack=True)

    # read the metadata
    if (no_metadata is False):
        try:
            # init
            found_site_uid = None

            # read file's metadata
            fp = open(file, 'r')
            for line in fp:
                # check if we want this line
                if (line[0] != '#'):
                    break
                if ("------------" in line):
                    # end of metadata, bail out
                    break
                line = line.strip()[1:]  # remove the whitespace, remove the # leading character

                # split the line based
                #
                # NOTE: some lines are differently formatted, so we
                # need a few special cases to handle them
                if ("Version" in line):  # version line
                    metadata_dict["version"] = line.strip()
                elif ("----" in line):  # first line
                    metadata_dict["summary"] = line.strip()
                else:
                    line_split = [x.strip() for x in line.split(':', maxsplit=1)]
                    if ("processing date" in line.lower()):  # convert to ISO datetime format
                        metadata_dict["processing_date"] = datetime.datetime.strptime(line_split[1],
                                                                                      "%a %b %d %H:%M:%S %Y").strftime("%Y-%m-%d %H:%M:%S")
                    elif ("site unique id" in line.lower()):  # site ID line
                        # skip this line since it is not consistently in the metadata, so
                        # we will derive it from the filename and insert it after this for-loop
                        found_site_uid = line_split[1].lower()
                        continue
                    else:
                        metadata_dict[line_split[0].lower().replace(' ', '_')] = line_split[1]
            fp.close()

            # add in the site unique ID if it isn't in there
            if (found_site_uid is not None):  # already have it
                metadata_dict["site_unique_id"] = found_site_uid
            else:
                if ('_' == file.name[3]):
                    # 3-letter site code
                    for s4, s3 in NORSTAR_RIOMETER_3_LETTER_SITE_CODES.items():
                        if (file.name[0:3] in s3):
                            metadata_dict["site_unique_id"] = s4
                            break
                else:
                    # 4-letter site code -- find position of 'rio-' in the file, site code
                    # will be right after that
                    idx = file.name.find('rio-')
                    if (idx != -1):
                        metadata_dict["site_unique_id"] = file[idx + 4:4].lower()
            if ("site_unique_id" not in metadata_dict):
                # still haven't found it, raise a warning message
                metadata_dict["site_unique_id"] = "unknown"
                if (quiet is False):
                    print("Warning: unable to determine 'site_unique_id' field in file '%s'" % (file))
        except Exception as e:
            # set error message
            if (quiet is False):
                print("Error reading metadata for file '%s': %s" % (file, str(e)))
            problematic = True
            error_message = "error reading metadata: %s" % (str(e))

            # reset objects
            metadata_dict = {}
            np_raw_signal = np.array([], dtype=RIOMETER_DT)
            np_absorption = None if file_type == "k0" else np.array([], dtype=RIOMETER_DT)

            # return
            return {
                "np_timestamp": np_timestamp,
                "np_raw_signal": np_raw_signal,
                "np_absorption": np_absorption,
                "metadata": metadata_dict,
                "problematic": problematic,
                "file": file,
                "error_message": error_message,
            }

    # set timestamp numpy array
    try:
        # initialize new object, populate using the date and time arrays
        #
        # NOTE: there can be UT24 records in the files, so we need to keep track
        # of those, and delete them from the raw_signal and absorption arrays
        #
        # NOTE: this is also where we filter based on the start and end times
        idxs_to_delete = []
        np_timestamp = np.empty(np_date.shape, dtype=datetime.datetime)
        for i in range(0, np_date.shape[0]):
            # mark the 24th hour items as indexes to delete
            if (np_time[i][0:2].decode() == "24"):
                idxs_to_delete.append(i)
                continue

            # mark times outside of the valid start and end ranges as indexes to delete
            np_timestamp[i] = datetime.datetime.strptime("%s %s" % (np_date[i].decode(), np_time[i].decode()), "%d/%m/%y %H:%M:%S")
            if ((start_time is None or np_timestamp[i] >= start_time) and (end_time is None or np_timestamp[i] <= end_time)):
                pass
            else:
                idxs_to_delete.append(i)

        # delete rows as needed
        if (len(idxs_to_delete) > 0):
            np_timestamp = np.delete(np_timestamp, idxs_to_delete)
            np_raw_signal = np.delete(np_raw_signal, idxs_to_delete)
            if (np_absorption is not None):
                np_absorption = np.delete(np_absorption, idxs_to_delete)
    except Exception as e:
        if (quiet is False):
            print("Error processing timestamps for file '%s': %s" % (file, str(e)))
        problematic = True
        error_message = "error processing timestamps: %s" % (str(e))

        # set
        metadata_dict = {}
        np_timestamp = np.array([], dtype=datetime.datetime)
        np_raw_signal = np.array([], dtype=RIOMETER_DT)
        np_absorption = None if file_type == "k0" else np.array([], dtype=RIOMETER_DT)

        # return
        return {
            "np_timestamp": np_timestamp,
            "np_raw_signal": np_raw_signal,
            "np_absorption": np_absorption,
            "metadata": metadata_dict,
            "problematic": problematic,
            "file": file,
            "error_message": error_message,
        }

    # return
    return {
        "np_timestamp": np_timestamp,
        "np_raw_signal": np_raw_signal,
        "np_absorption": np_absorption,
        "metadata": metadata_dict,
        "problematic": problematic,
        "file": file,
        "error_message": error_message,
    }



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Functions for running TREx Auroral Transport Model (ATM) calculations.
"""

import datetime
from .classes_forward import *
from .classes_inverse import *
from ._forward import forward as func_forward
from ._inverse import inverse as func_inverse

ATM_DEFAULT_MAXWELLIAN_ENERGY_FLUX = 10.0
ATM_DEFAULT_GAUSSIAN_ENERGY_FLUX = 0.0
ATM_DEFAULT_MAXWELLIAN_CHARACTERISTIC_ENERGY = 5000.0
ATM_DEFAULT_GAUSSIAN_PEAK_ENERGY = 1000.0
ATM_DEFAULT_GAUSSIAN_SPECTRAL_WIDTH = 100.0
ATM_DEFAULT_NRLMSIS_MODEL_VERSION = "2.0"
ATM_DEFAULT_OXYGEN_CORRECTION_FACTOR = 1.0
ATM_DEFAULT_TIMESCALE_AURORAL = 600
ATM_DEFAULT_TIMESCALE_TRANSPORT = 600
ATM_DEFAULT_MODEL_VERSION = "1.0"
ATM_DEFAULT_PRECIPITATION_SPECTRAL_FLUX_TYPE = "gaussian"


class ATMManager:
    """
    The ATMManager object is initialized within every PyUCalgarySRS.models object. It 
    acts as a way to access the submodules and carry over configuration information in 
    the super class.
    """

    def __init__(self, srs_obj):
        self.__srs_obj = srs_obj

    def forward(self,
                timestamp: datetime.datetime,
                geodetic_latitude: float,
                geodetic_longitude: float,
                output: ATMForwardOutputFlags,
                maxwellian_energy_flux: float = ATM_DEFAULT_MAXWELLIAN_ENERGY_FLUX,
                gaussian_energy_flux: float = ATM_DEFAULT_GAUSSIAN_ENERGY_FLUX,
                maxwellian_characteristic_energy: float = ATM_DEFAULT_MAXWELLIAN_CHARACTERISTIC_ENERGY,
                gaussian_peak_energy: float = ATM_DEFAULT_GAUSSIAN_PEAK_ENERGY,
                gaussian_spectral_width: float = ATM_DEFAULT_GAUSSIAN_SPECTRAL_WIDTH,
                nrlmsis_model_version: Literal["00", "2.0"] = ATM_DEFAULT_NRLMSIS_MODEL_VERSION,
                oxygen_correction_factor: float = ATM_DEFAULT_OXYGEN_CORRECTION_FACTOR,
                timescale_auroral: int = ATM_DEFAULT_TIMESCALE_AURORAL,
                timescale_transport: int = ATM_DEFAULT_TIMESCALE_TRANSPORT,
                atm_model_version: Literal["1.0"] = ATM_DEFAULT_MODEL_VERSION,
                custom_spectrum: Optional[ndarray] = None,
                no_cache: bool = False,
                timeout: Optional[int] = None) -> ATMForwardResult:
        """
        Perform a forward calculation using the TREx Auroral Transport Model and the supplied input 
        parameters. Note that this function utilizes the UCalgary Space Remote Sensing API to perform 
        the calculation.

        The ATM model is 1D and time-independent. However, the optional parameters `timescale_auroral` 
        and `timescale_transport` provide limited support for time-dependent and transport process. The
        `timescale_auroral` parameter (T0) is the duration of the precipitation. The `timescale_transport` 
        parameter is defined by L/v0, in which L is the dimension of the auroral structure, and v0 is the 
        cross-structure drift speed. The model quasi-analytically solves the continuity equation under a 
        square input (with time duration T0 and spatial width L) input of precipitation. The initial/boundary 
        conditions are given by IRI. The output yields the mean density/VER over [0-L] at time T0.

        Args:
            timestamp (datetime.datetime): 
                Timestamp for the calculation. This value is expected to be in UTC, and is valid for 
                any value up to the end of the previous day. Any timezone data will be ignored. This 
                parameter is required.

            geodetic_latitude (float): 
                Latitude in geodetic coordinates: -90.0 to 90.0. This parameter is required.

            geodetic_longitude (float): 
                Longitude in geodetic coordinates: -180.0 to 180.0. This parameter is required.

            output (ATMForwardOutputFlags): 
                Flags to indicate which values are included in the output. See 
                [`ATMForwardOutputFlags`](classes_forward.html#pyucalgarysrs.models.atm.classes_forward.ATMForwardOutputFlags) 
                for more details. This parameter is required.

            maxwellian_energy_flux (float): 
                Maxwellian energy flux in erg/cm2/s. Default is 10. This parameter is optional.

            gaussian_energy_flux (float): 
                Gaussian energy flux in erg/cm2/s. Default is 0.0. Note that `gaussian_peak_energy` and
                `gaussian_spectral_width` must be specified if the `gaussian_energy_flux` is not 0. 
                This parameter is optional.

            maxwellian_characteristic_energy (float): 
                Maxwellian characteristic energy in eV. Default is 5000. Note that `maxwellian_characteristic_energy` 
                must be specified if the `maxwellian_energy_flux` is not 0. This parameter 
                is optional.

            gaussian_peak_energy (float): 
                Gaussian peak energy in eV. Default is 1000. Note this parameter must be specified if 
                the `gaussian_energy_flux` is not 0. This parameter is optional.

            gaussian_spectral_width (float): 
                Gaussian spectral width in eV. Default is 100. Note this parameter must be specified if 
                the `gaussian_energy_flux` is not 0. This parameter is optional.

            nrlmsis_model_version (str): 
                NRLMSIS version number. Possible values are `00` or `2.0`. Default is `2.0`. This parameter is
                optional. More details about this empirical model can be found [here](https://ccmc.gsfc.nasa.gov/models/NRLMSIS~00/),
                and [here](https://ccmc.gsfc.nasa.gov/models/NRLMSIS~2.0/).

            oxygen_correction_factor (float): 
                Oxygen correction factor used to multiply by in the empirical model. Default is 1. This parameter
                is optional.

            timescale_auroral (int): 
                The duration of the precipitation, in seconds. Default is 600 (10 minutes). This parameter is optional.

            timescale_transport (int): 
                Defined by L/v0, in which L is the dimension of the auroral structure, and v0 is the cross-structure drift 
                speed. Represented in seconds. Default is 600 (10 minutes). This parameter is optional.

            atm_model_version (str): 
                ATM model version number. Possible values are only '1.0' at this time, but will have
                additional possible values in the future. This parameter is optional.

            custom_spectrum (ndarray): 
                A 2-dimensional numpy array (dtype is any float type) containing values representing the
                energy in eV, and flux in 1/cm2/sr/eV. The shape is expected to be [N, 2], with energy in
                [:, 0] and flux in [:, 1]. Note that this array cannot contain negative values (SRSAPIError 
                will be raised if so). This parameter is optional.

            no_cache (bool): 
                The UCalgary Space Remote Sensing API utilizes a caching layer for performing ATM
                calculations. If this variation of input parameters has been run before (and the
                cache is still valid), then it will not re-run the calculation. Instead it will 
                return the cached results immediately. To disable the caching layer, set this 
                parameter to `True`. Default is `False`. This parameter is optional.

            timeout (int): 
                Represents how many seconds to wait for the API to send data before giving up. The 
                default is 30 seconds, or the `api_timeout` value in the super class' `pyucalgarysrs.PyUCalgarySRS`
                object. This parameter is optional.
        
        Returns:
            An [`ATMForwardResult`](classes_forward.html#pyucalgarysrs.models.atm.classes_forward.ATMForwardResult)
            object containing the requested output data, among other values.

        Raises:
            pyucalgarysrs.exceptions.SRSAPIError: An API error was encountered
        """
        return func_forward(
            self.__srs_obj,
            timestamp,
            geodetic_latitude,
            geodetic_longitude,
            output,
            maxwellian_energy_flux,
            gaussian_energy_flux,
            maxwellian_characteristic_energy,
            gaussian_peak_energy,
            gaussian_spectral_width,
            nrlmsis_model_version,
            oxygen_correction_factor,
            timescale_auroral,
            timescale_transport,
            atm_model_version,
            custom_spectrum,
            no_cache,
            timeout,
        )

    def inverse(self,
                timestamp: datetime.datetime,
                geodetic_latitude: float,
                geodetic_longitude: float,
                intensity_4278: float,
                intensity_5577: float,
                intensity_6300: float,
                intensity_8446: float,
                output: ATMInverseOutputFlags,
                precipitation_flux_spectral_type: Literal["gaussian", "maxwellian"] = ATM_DEFAULT_PRECIPITATION_SPECTRAL_FLUX_TYPE,
                nrlmsis_model_version: Literal["00", "2.0"] = ATM_DEFAULT_NRLMSIS_MODEL_VERSION,
                atmospheric_attenuation_correction: bool = False,
                atm_model_version: Literal["1.0"] = ATM_DEFAULT_MODEL_VERSION,
                no_cache: bool = False,
                timeout: Optional[int] = None) -> ATMInverseResult:
        """
        Perform an inverse calculation using the TREx Auroral Transport Model and the supplied input 
        parameters. Note that this function utilizes the UCalgary Space Remote Sensing API to perform 
        the calculation.

        Args:
            timestamp (datetime.datetime): 
                Timestamp for the calculation. This value is expected to be in UTC, and is valid for a pre-defined 
                timeframe. An error will be raised if outside of the valid timeframe. Any timezone data will be 
                ignored. This parameter is required.

            geodetic_latitude (float): 
                Latitude in geodetic coordinates. Currently limited to the Transition Region Explorer (TREx)
                region of >=50.0 and <61.5 degrees. An error will be raised if outside of this range. This 
                parameter is required.

            geodetic_longitude (float): 
                Longitude in geodetic coordinates. Currently limited to the Transition Region Explorer (TREx)
                region of >=-110 and <-70 degrees. An error will be raised if outside of this range. This 
                parameter is required.

            intensity_4278 (float): 
                Intensity of the 427.8nm (blue) wavelength in Rayleighs. This parameter is required.                

            intensity_5577 (float): 
                Intensity of the 557.7nm (green) wavelength in Rayleighs. This parameter is required.                

            intensity_6300 (float): 
                Intensity of the 630.0nm (red) wavelength in Rayleighs. This parameter is required.                

            intensity_8446 (float): 
                Intensity of the 844.6nm (near infrared) wavelength in Rayleighs. This parameter is required.                

            output (ATMInverseOutputFlags): 
                Flags to indicate which values are included in the output. See 
                [`ATMInverseOutputFlags`](classes_inverse.html#pyucalgarysrs.models.atm.classes_inverse.ATMInverseOutputFlags) 
                for more details. This parameter is required.

            precipitation_flux_spectral_type (str): 
                The precipitation flux spectral type to use. Possible values are `gaussian` or `maxwellian`. The
                default is `gaussian`. This parameter is optional.

            nrlmsis_model_version (str): 
                NRLMSIS version number. Possible values are `00` or `2.0`. Default is `2.0`. This parameter is
                optional. More details about this empirical model can be found [here](https://ccmc.gsfc.nasa.gov/models/NRLMSIS~00/),
                and [here](https://ccmc.gsfc.nasa.gov/models/NRLMSIS~2.0/).

            atmospheric_attenuation_correction (bool): 
                Apply an atmospheric attenuation correction factor. Default is `False`.

            atm_model_version (str): 
                ATM model version number. Possible values are only '1.0' at this time, but will have
                additional possible values in the future. This parameter is optional.

            no_cache (bool): 
                The UCalgary Space Remote Sensing API utilizes a caching layer for performing ATM
                calculations. If this variation of input parameters has been run before (and the
                cache is still valid), then it will not re-run the calculation. Instead it will 
                return the cached results immediately. To disable the caching layer, set this 
                parameter to `True`. Default is `False`. This parameter is optional.

            timeout (int): 
                Represents how many seconds to wait for the API to send data before giving up. The 
                default is 30 seconds, or the `api_timeout` value in the super class' `pyucalgarysrs.PyUCalgarySRS`
                object. This parameter is optional.

        Returns:
            An [`ATMInverseResult`](classes_inverse.html#pyucalgarysrs.models.atm.classes_inverse.ATMInverseResult)
            object containing the requested output data, among other values.

        Raises:
            pyucalgarysrs.exceptions.SRSAPIError: An API error was encountered
        """
        return func_inverse(
            self.__srs_obj,
            timestamp,
            geodetic_latitude,
            geodetic_longitude,
            intensity_4278,
            intensity_5577,
            intensity_6300,
            intensity_8446,
            output,
            precipitation_flux_spectral_type,
            nrlmsis_model_version,
            atmospheric_attenuation_correction,
            atm_model_version,
            no_cache,
            timeout,
        )



# Copyright 2024 University of Calgary
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import signal
import numpy as np
import h5py
from pathlib import Path
from multiprocessing import Pool
from functools import partial
from ..classes import HSRData

# globals
HSR_DT = np.dtype("float32")


def read(file_list, n_parallel=1, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # if input is just a single file name in a string, convert to a list to be fed to the workers
    if isinstance(file_list, str) or isinstance(file_list, Path):
        file_list = [file_list]

    # check n_parallel
    if (n_parallel > 1):
        try:
            # set up process pool (ignore SIGINT before spawning pool so child processes inherit SIGINT handler)
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            pool = Pool(processes=n_parallel)
            signal.signal(signal.SIGINT, original_sigint_handler)  # restore SIGINT handler
        except ValueError:  # pragma: nocover
            # likely the read call is being used within a context that doesn't support the usage
            # of signals in this way, proceed without it
            pool = Pool(processes=n_parallel)

        # call readfile function, run each iteration with a single input file from file_list
        # NOTE: structure of data - data[file][metadata dictionary lists = 1, images = 0][frame]
        data = []
        try:
            data = pool.map(partial(
                __riometer_readfile_worker,
                no_metadata=no_metadata,
                start_time=start_time,
                end_time=end_time,
                quiet=quiet,
            ), file_list)
        except KeyboardInterrupt:  # pragma: nocover
            pool.terminate()  # gracefully kill children
            return [], [], [], []
        else:
            pool.close()
            pool.join()
    else:
        # don't bother using multiprocessing with one worker, just call the worker function directly
        data = []
        for f in file_list:
            data.append(__riometer_readfile_worker(
                f,
                no_metadata=no_metadata,
                start_time=start_time,
                end_time=end_time,
                quiet=quiet,
            ))

    # compile results
    hsr_data = []
    top_level_timestamps = []
    metadata = []
    problematic_file_list = []
    for result in data:
        # problematic
        if (result["problematic"] is True):
            problematic_file_list.append({
                "filename": result["file"],
                "error_message": result["error_message"],
            })
            continue

        # hsr data
        hsr_data.append(
            HSRData(
                band_central_frequency=result["band_central_frequency_list"],
                band_passband=result["band_passband_list"],
                timestamp=result["np_timestamp"],
                raw_power=result["np_raw_power"],
                absorption=result["np_absorption"],
            ))

        # timestamps
        if (len(result["np_timestamp"]) > 0):
            top_level_timestamps.append(result["np_timestamp"][0])

        # metadata
        if (no_metadata is False):
            metadata.append(result["metadata"])

    # return
    return hsr_data, top_level_timestamps, metadata, problematic_file_list


def __str_to_datetime_formatter(timestamp_str):
    return datetime.datetime.strptime(timestamp_str.decode(), "%Y-%m-%d %H:%M:%S UTC")


def __riometer_readfile_worker(file, no_metadata=False, start_time=None, end_time=None, quiet=False):
    # init
    metadata_dict = {}
    band_central_frequency_list = []
    band_passband_list = []
    np_timestamp = np.array([], dtype=datetime.datetime)
    np_raw_power = np.array([], dtype=HSR_DT)
    np_absorption = np.array([], dtype=HSR_DT)
    problematic = False
    error_message = ""

    # process file
    try:
        # open H5 file
        f = h5py.File(file, 'r')

        # get timestamp, convert into datetime objects
        np_timestamp = f["data"]["timestamp"][:]  # type: ignore
        np_timestamp = np.vectorize(__str_to_datetime_formatter)(np_timestamp)  # type: ignore
        np_timestamp = np_timestamp.astype(datetime.datetime)

        # determine indexes to keep based on start and end times
        idxs = []
        for idx, val in enumerate(np_timestamp):
            if ((start_time is None or val >= start_time) and (end_time is None or val <= end_time)):
                idxs.append(idx)
        np_timestamp = np_timestamp[idxs]

        # get raw power, timestamp
        np_raw_power = f["data"]["raw_power"][:, idxs]  # type: ignore

        # get band central frequency, bandpass
        band_central_frequency_list = f["data"]["band_central_frequency"][:].tolist()  # type: ignore
        band_passband_list = f["data"]["band_passband"][:].tolist()  # type: ignore

        # set absorption
        if ("_k0_" in file.name):
            # k0 data, no absorption
            np_absorption = None

        # get metadata
        if (no_metadata is False):
            for key, value in f["metadata"]["file"].attrs.items():  # type: ignore
                metadata_dict[key] = value

        # close file
        f.close()
    except Exception as e:
        # set error message
        if (quiet is False):
            print("Failed to read file '%s': %s" % (file, str(e)))
        problematic = True
        error_message = "failed to open file: %s" % (str(e))

        # clear values
        band_central_frequency_list = []
        band_passband_list = []
        np_timestamp = np.array([], dtype=datetime.datetime)
        np_raw_power = np.array([], dtype=HSR_DT)
        np_absorption = np.array([], dtype=HSR_DT)

        # close file
        try:
            f.close()  # type: ignore
        except Exception:
            pass

        # return
        return {
            "band_central_frequency_list": band_central_frequency_list,
            "band_passband_list": band_passband_list,
            "np_timestamp": np_timestamp,
            "np_raw_power": np_raw_power,
            "np_absorption": np_absorption,
            "metadata": metadata_dict,
            "problematic": problematic,
            "file": file,
            "error_message": error_message,
        }

    # return
    return {
        "band_central_frequency_list": band_central_frequency_list,
        "band_passband_list": band_passband_list,
        "np_timestamp": np_timestamp,
        "np_raw_power": np_raw_power,
        "np_absorption": np_absorption,
        "metadata": metadata_dict,
        "problematic": problematic,
        "file": file,
        "error_message": error_message,
    }
